# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import re

from six import integer_types, string_types

from efc.rpn_builder.parser.operands import (CellRangeOperand, CellSetOperand, ErrorOperand, NotFoundErrorOperand,
                                             RPNOperand, SimpleOperand, SingleCellOperand, ValueErrorOperand)

__all__ = ('EXCEL_FUNCTIONS',)


def _get_type_id(obj):
    if isinstance(obj, bool):
        return 2
    elif isinstance(obj, string_types):
        return 1
    return 0


def type_mixin(a, b):
    if a is None:
        a = '' if isinstance(b, string_types) else 0
    if b is None:
        b = '' if isinstance(a, string_types) else 0
    return (_get_type_id(a), a), (_get_type_id(b), b)


def add_func(*args):
    if len(args) == 2:
        op1, op2 = args
        return op1.digit + op2.digit
    else:
        return args[0].digit


def subtract_func(*args):
    if len(args) == 2:
        op1, op2 = args
        return op1.digit - op2.digit
    else:
        return -args[0].digit


def divide_func(op1, op2):
    return op1.digit / op2.digit


def multiply_func(op1, op2):
    return op1.digit * op2.digit


def concat_func(op1, op2):
    a = op1.value
    b = op2.value
    if isinstance(a, float) and a % 1 == 0:
        a = int(a)
    if isinstance(b, float) and b % 1 == 0:
        b = int(b)
    return '%s%s' % (a, b)


def exponent_func(op1, op2):
    return op1.digit ** op2.digit


def compare_not_eq_func(op1, op2):
    op1, op2 = type_mixin(op1.value, op2.value)
    return op1 != op2


def compare_gte_func(op1, op2):
    op1, op2 = type_mixin(op1.value, op2.value)
    return op1 >= op2


def compare_lte_func(op1, op2):
    op1, op2 = type_mixin(op1.value, op2.value)
    return op1 <= op2


def compare_gt_func(op1, op2):
    op1, op2 = type_mixin(op1.value, op2.value)
    return op1 > op2


def compare_lt_func(op1, op2):
    op1, op2 = type_mixin(op1.value, op2.value)
    return op1 < op2


def compare_eq_func(op1, op2):
    op1, op2 = type_mixin(op1.value, op2.value)
    return op1 == op2


def iter_elements(*args):
    for arg in args:
        if isinstance(arg, RPNOperand):
            arg = arg.evaluated_value

        if isinstance(arg, (CellRangeOperand, CellSetOperand)):
            for cell in arg:
                yield cell
        else:
            yield arg


def iter_digits(yield_none, *args):
    for op in iter_elements(*args):
        if op.is_blank:
            if yield_none:
                yield None
        else:
            try:
                yield op.digit
            except:
                pass


def sum_func(*args):
    return sum(iter_digits(False, *args))


def mod_func(op1, op2):
    return op1.digit % op2.digit


def if_func(expr_op, op1, op2=None):
    if op2 is None:
        op2 = False
    return op1 if expr_op.value else op2


def if_error_func(op1, op2):
    if isinstance(op1, RPNOperand):
        op1 = op1.evaluated_value
    return op2 if isinstance(op1, ErrorOperand) else op1


def max_func(*args):
    return max(list(iter_digits(False, *args)) or [0])


def min_func(*args):
    return min(list(iter_digits(False, *args)) or [0])


def left_func(op1, op2=1):
    return op1.string[:int(op2)]


def right_func(op1, op2=1):
    return op1.string[-int(op2):]


def mid_func(op1, op2, op3):
    left = int(op2) - 1
    right = left + int(op3)
    return op1.string[left:right]


def is_blank_func(a):
    return a.value is None


def or_function(*args):
    for op in iter_elements(*args):
        v = op.value
        if v is not None and not isinstance(v, string_types) and v:
            return True
    return False


def and_function(*args):
    for op in iter_elements(*args):
        v = op.value
        if v is not None and not isinstance(v, string_types) and not v:
            return False
    return True


def not_func(op):
    return not op.value


def small_function(r, op):
    items = sorted(iter_digits(False, r))
    index = int(op) - 1
    try:
        return items[index]
    except IndexError:
        return ValueErrorOperand()


def large_function(r, op):
    items = sorted(iter_digits(False, r), reverse=True)
    index = int(op) - 1
    try:
        return items[index]
    except IndexError:
        return ValueErrorOperand()


def round_function(a, b):
    return round(a.digit, int(b))


def round_down_function(a, b):
    base = 10 ** (int(b))
    return a.digit * base // 1 / base


def floor_function(a, multiple):
    multiple = int(multiple)
    return int(a.digit / multiple) * multiple


def count_function(*args):
    return len([op for op in iter_elements(*args)
                if not op.is_blank and isinstance(op.value, (integer_types, float))])


def abs_function(a):
    return abs(a.digit)


def match_function(op1, r, match_type=None):
    match_type = 0 if match_type is None else int(match_type)

    expr = op1.value
    idx = None
    if match_type == 1:
        for idx, item in enumerate(r, 1):
            a, b = type_mixin(item.value, expr)
            if a == b:
                break
            elif a > b:
                idx -= 1
                break
    elif match_type == -1:
        for idx, item in enumerate(r, 1):
            a, b = type_mixin(item.value, expr)
            if a > b:
                break
        else:
            idx = None
    else:
        for idx, item in enumerate(r, 1):
            a, b = type_mixin(item.value, expr)
            if a == b:
                break
        else:
            idx = None
    if idx is None:
        raise NotFoundErrorOperand()
    return idx


COUNT_IF_EXPR = re.compile(r'^(?P<symbol><=|>=|<>|>|<|=)(?P<value>.+)$')


def get_check_function(expr):
    if isinstance(expr.value, string_types):
        match = COUNT_IF_EXPR.search(expr.value)
        if match:
            match = match.groupdict()
            operation = match['symbol']
            operand = match['value']
        else:
            operation = '='
            operand = expr.value
    else:
        operation = '='
        operand = expr.value

    try:
        operand = float(operand)
    except (TypeError, ValueError):
        pass

    check = ARITHMETIC_FUNCTIONS[operation]
    return check, SimpleOperand(operand)


def countif_function(cells, expr):
    check, operand = get_check_function(expr)
    return len([op for op in cells.value if not op.is_blank and check(op, operand)])


def ifs_indexes(*args):
    args = iter(args)
    good_indexes = set()
    first_iteration = True
    while True:
        try:
            op = next(args)
        except StopIteration:
            break

        check, expr = get_check_function(next(args))
        for idx, item in enumerate(op, 1):
            if check(item, expr):
                if first_iteration:
                    good_indexes.add(idx)
            elif idx in good_indexes:
                good_indexes.remove(idx)
        first_iteration = False
    return good_indexes


def sum_ifs_function(op1, *args):
    good_indexes = ifs_indexes(*args)
    return sum_func(*[c for idx, c in enumerate(op1, 1) if idx in good_indexes])


def concatenate(*args):
    return ''.join(i.string for i in iter_elements(*args) if not i.is_blank)


def average_function(*args):
    values = list(iter_digits(False, *args))
    return sum(values) / len(values)


def average_ifs_function(op1, *args):
    good_indexes = ifs_indexes(*args)
    return average_function(*[c for idx, c in enumerate(op1, 1) if idx in good_indexes])


def count_blank_function(cells):
    return len([op for op in iter_elements(cells) if op.is_blank])


def count_ifs_function(op1, *args):
    good_indexes = ifs_indexes(*args)
    return count_function(*[c for idx, c in enumerate(op1, 1) if idx in good_indexes])


def offset_function(cell, row_offset, col_offset, height=None, width=None):
    if not isinstance(cell, SingleCellOperand):
        return ValueErrorOperand()

    height = int(height) if height is not None else 1
    width = int(width) if width is not None else 1

    if height == width == 1:
        return SingleCellOperand(row=cell.row + int(row_offset), column=cell.column + int(col_offset),
                                 ws_name=cell.ws_name, source=cell.source)
    else:
        return CellRangeOperand(row1=cell.row + int(row_offset),
                                column1=cell.column + int(col_offset),
                                row2=cell.row + int(row_offset) + height - 1,
                                column2=cell.column + int(col_offset) + width - 1,
                                ws_name=cell.ws_name, source=cell.source)


def vlookup_function(op, rg, column, flag=None):
    first_col = rg.offset()
    first_col.column2 = first_col.column1

    if flag is not None and flag.digit or flag is None:
        idx = match_function(op, first_col, 1)
        if flag is not None and flag.digit and idx != 1:
            a, b = type_mixin(list(first_col)[idx - 1].value, op.value)
            if a == b:
                idx -= 1
    else:
        idx = match_function(op, first_col, 0)
    return SingleCellOperand(row=(rg.row1 or 1) + idx - 1, column=(rg.column1 or 1) + column.digit - 1,
                             ws_name=rg.ws_name, source=rg.source)


ARITHMETIC_FUNCTIONS = {
    '+': add_func,
    '-': subtract_func,
    '/': divide_func,
    '*': multiply_func,
    '&': concat_func,
    '^': exponent_func,
    '<>': compare_not_eq_func,
    '>=': compare_gte_func,
    '<=': compare_lte_func,
    '>': compare_gt_func,
    '<': compare_lt_func,
    '=': compare_eq_func,
}

EXCEL_FUNCTIONS = {}
EXCEL_FUNCTIONS.update(ARITHMETIC_FUNCTIONS)

EXCEL_FUNCTIONS['SUM'] = sum_func
EXCEL_FUNCTIONS['SUMIFS'] = sum_ifs_function
EXCEL_FUNCTIONS['MOD'] = mod_func
EXCEL_FUNCTIONS['IF'] = if_func
EXCEL_FUNCTIONS['IFERROR'] = if_error_func
EXCEL_FUNCTIONS['MAX'] = max_func
EXCEL_FUNCTIONS['MIN'] = min_func
EXCEL_FUNCTIONS['LEFT'] = left_func
EXCEL_FUNCTIONS['RIGHT'] = right_func
EXCEL_FUNCTIONS['MID'] = mid_func
EXCEL_FUNCTIONS['ISBLANK'] = is_blank_func
EXCEL_FUNCTIONS['OR'] = or_function
EXCEL_FUNCTIONS['AND'] = and_function
EXCEL_FUNCTIONS['NOT'] = not_func
EXCEL_FUNCTIONS['ROUND'] = round_function
EXCEL_FUNCTIONS['ROUNDDOWN'] = round_down_function
EXCEL_FUNCTIONS['FLOOR'] = floor_function
EXCEL_FUNCTIONS['COUNT'] = count_function
EXCEL_FUNCTIONS['COUNTIF'] = countif_function
EXCEL_FUNCTIONS['COUNTBLANK'] = count_blank_function
EXCEL_FUNCTIONS['ABS'] = abs_function
EXCEL_FUNCTIONS['OFFSET'] = offset_function
EXCEL_FUNCTIONS['MATCH'] = match_function
EXCEL_FUNCTIONS['AVERAGE'] = average_function
EXCEL_FUNCTIONS['AVERAGEIFS'] = average_ifs_function
EXCEL_FUNCTIONS['VLOOKUP'] = vlookup_function
EXCEL_FUNCTIONS['SMALL'] = small_function
EXCEL_FUNCTIONS['LARGE'] = large_function
EXCEL_FUNCTIONS['COUNTIFS'] = count_ifs_function
EXCEL_FUNCTIONS['CONCATENATE'] = concatenate



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import re

import pytest

from efc.rpn_builder.lexer import tokens
from efc.utils import col_str_to_index


@pytest.mark.parametrize(
    ['line', 'token_type'],
    (
            ('+', tokens.AddToken),
            ('-', tokens.SubtractToken),
            ('/', tokens.DivideToken),
            ('*', tokens.MultiplyToken),
            ('&', tokens.ConcatToken),
            ('^', tokens.ExponentToken),
            ('<>', tokens.CompareNotEqToken),
            ('>=', tokens.CompareGTEToken),
            ('<=', tokens.CompareLTEToken),
            ('>', tokens.CompareGTToken),
            ('<', tokens.CompareLTToken),
            ('=', tokens.CompareEqToken),
    )
)
def test_arithmetic_tokens(line, token_type):
    match = re.match('(?P<%s>%s)' % (token_type.__name__, token_type.pattern), line)
    token = token_type(match)
    assert token.src_value == line
    assert token.token_value == line


def test_function_token():
    pattern = re.compile('(?P<%s>%s)' % (tokens.FunctionToken.__name__, tokens.FunctionToken.pattern))

    token = tokens.FunctionToken(pattern.match('SUM('))
    assert token.src_value == 'SUM'
    assert token.token_value == 'SUM'

    assert pattern.match('SUM') is None


@pytest.mark.parametrize(
    ['line', 'token_type', 'result'],
    (
            ('1', tokens.IntToken, 1),
            ('1234', tokens.IntToken, 1234),
            ('1.0', tokens.FloatToken, 1.0),
            ('111.0', tokens.FloatToken, 111.0),
            ('1.001', tokens.FloatToken, 1.001),
            ('TRUE', tokens.BoolToken, True),
            ('FALSE', tokens.BoolToken, False),
            ('"FALSE"', tokens.StringToken, 'FALSE'),
            ('"hello"', tokens.StringToken, 'hello'),
            ('"HELLO"', tokens.StringToken, 'HELLO'),
    )
)
def test_simple_type_tokens(line, token_type, result):
    match = re.match('(?P<%s>%s)' % (token_type.__name__, token_type.pattern), line)
    token = token_type(match)
    assert token.src_value == line
    assert token.token_value == result


def gen_address_ws_name(doc, ws_name):
    name_list = []
    if doc is not None:
        name_list.append('[%s]' % doc)

    if ws_name is not None:
        name_list.append('%s' % ws_name)

    if doc is not None or ws_name is not None:
        name_list.append('!')
    return ''.join(name_list)


def gen_cell(row, row_fixed, col, col_fixed):
    col = ('$%s' % col if col_fixed else col) if col is not None else ''
    row = ('$%s' % row if row_fixed else row) if row is not None else ''
    return col + row


@pytest.mark.parametrize('doc', [None, 'src'])
@pytest.mark.parametrize('ws_name', [None, 'word', '1_word', '\'two words\''])
@pytest.mark.parametrize('row', ['1', '100'])
@pytest.mark.parametrize('column', ['A', 'AA'])
@pytest.mark.parametrize('row_fixed', [True, False])
@pytest.mark.parametrize('column_fixed', [True, False])
def test_single_cell_token(doc, ws_name, row, column, row_fixed, column_fixed):
    address = gen_address_ws_name(doc, ws_name) + gen_cell(row, row_fixed, column, column_fixed)

    match = re.match('(?P<%s>%s)' % (tokens.SingleCellToken.__name__, tokens.SingleCellToken.pattern), address)
    token = tokens.SingleCellToken(match)
    if ws_name is None:
        assert token.ws_name is None
    else:
        assert token.ws_name == ws_name.replace('\'', '')
    assert token.row == int(row)
    assert token.column == col_str_to_index(column)
    assert token.row_fixed == row_fixed
    assert token.column_fixed == column_fixed

    with pytest.raises(KeyError):
        token.s_doc == doc


@pytest.mark.parametrize('doc', [None, 'src'])
@pytest.mark.parametrize('ws_name', [None, 'word', '1_word', '\'two words\''])
@pytest.mark.parametrize('row1', ['1', '100'])
@pytest.mark.parametrize('column1', ['A', 'AA'])
@pytest.mark.parametrize('row1_fixed', [True, False])
@pytest.mark.parametrize('column1_fixed', [True, False])
@pytest.mark.parametrize('row2', ['1', '100'])
@pytest.mark.parametrize('column2', ['A', 'AA'])
@pytest.mark.parametrize('row2_fixed', [True, False])
@pytest.mark.parametrize('column2_fixed', [True, False])
def test_cells_range_token(doc, ws_name,
                           row1, column1, row1_fixed, column1_fixed,
                           row2, column2, row2_fixed, column2_fixed):
    cells_range = '%s:%s' % (gen_cell(row1, row1_fixed, column1, column1_fixed),
                             gen_cell(row2, row2_fixed, column2, column2_fixed))
    address = gen_address_ws_name(doc, ws_name) + cells_range

    match = re.match('(?P<%s>%s)' % (tokens.CellsRangeToken.__name__, tokens.CellsRangeToken.pattern), address)
    token = tokens.CellsRangeToken(match)
    if ws_name is None:
        assert token.ws_name is None
    else:
        assert token.ws_name == ws_name.replace('\'', '')
    assert token.row1 == int(row1)
    assert token.column1 == col_str_to_index(column1)
    assert token.row1_fixed == row1_fixed
    assert token.column1_fixed == column1_fixed

    assert token.row2 == int(row2)
    assert token.column2 == col_str_to_index(column2)
    assert token.row2_fixed == row2_fixed
    assert token.column2_fixed == column2_fixed

    with pytest.raises(KeyError):
        token.s_doc == doc


@pytest.mark.parametrize('doc', [None, 'src'])
@pytest.mark.parametrize('ws_name', [None, 'word', '1_word', '\'two words\''])
@pytest.mark.parametrize('range_name', ['name', '1_name'])
def test_named_range_token(doc, ws_name, range_name):
    address = gen_address_ws_name(doc, ws_name) + range_name

    match = re.match('(?P<%s>%s)' % (tokens.NamedRangeToken.__name__, tokens.NamedRangeToken.pattern), address)
    token = tokens.NamedRangeToken(match)
    if ws_name is None:
        assert token.ws_name is None
    else:
        assert token.ws_name == ws_name.replace('\'', '')
    assert token.name == range_name

    with pytest.raises(KeyError):
        token.s_doc == doc



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals


class BaseExcelInterface(object):
    """
    Base class to working with excel document
    """

    def cell_to_value(self, row, column, ws_name):
        """
        :type row: int
        :type column: int
        :type ws_name: basestring
        :rtype: list
        """
        raise NotImplementedError

    def named_range_to_cells(self, name, ws_name):
        """
        TEST_RANGE -> SingleCellOperand
        OTHER_RANGE -> CellRangeOperand
        OTHER2_RANGE -> CellSetOperand
        :type name: basestring
        :type ws_name: basestring
        :rtype: list
        """
        raise NotImplementedError

    def max_row(self, ws_name):
        raise NotImplementedError

    def min_row(self, ws_name):
        raise NotImplementedError

    def max_column(self, ws_name):
        raise NotImplementedError

    def min_column(self, ws_name):
        raise NotImplementedError

    def has_worksheet(self, ws_name):
        raise NotImplementedError

    def has_named_range(self, name, ws_name):
        raise NotImplementedError



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import pytest
from six.moves import zip

from efc.rpn_builder.lexer import Lexer
from efc.rpn_builder.parser import operands, operations, Parser


@pytest.fixture(scope='module')
def parse():
    lexer = Lexer()
    parser = Parser()
    return lambda line: parser.to_rpn(lexer.parse(line), None, None)


@pytest.mark.parametrize(
    ['line', 'token_types'],
    (
            ('4 + 5.54 - "hello"',
             [operands.SimpleOperand, operands.SimpleOperand, operands.SimpleOperand,
              operations.ArithmeticOperation, operations.ArithmeticOperation]),
            ('4 + 5.54 - SUM(1,2,4)',
             [operands.SimpleOperand, operands.SimpleOperand, operands.SimpleOperand, operands.SimpleOperand,
              operands.SimpleOperand, operations.FunctionOperation, operations.ArithmeticOperation,
              operations.ArithmeticOperation]),
            ('4 + 5.54 - SUM(1,SUM(1,2))',
             [operands.SimpleOperand, operands.SimpleOperand, operands.SimpleOperand, operands.RPNOperand,
              operations.FunctionOperation, operations.ArithmeticOperation, operations.ArithmeticOperation]),
            ('SUM(1,2,4) * 5',
             [operands.SimpleOperand, operands.SimpleOperand, operands.SimpleOperand, operations.FunctionOperation,
              operands.SimpleOperand, operations.ArithmeticOperation]),
            ('SUM(1 + 2,2,4) * 5',
             [operands.RPNOperand, operands.SimpleOperand,
              operands.SimpleOperand, operations.FunctionOperation, operands.SimpleOperand,
              operations.ArithmeticOperation]),
    )
)
def test_operations(parse, line, token_types):
    parsed_line = parse(line)
    assert len(parsed_line) == len(token_types), 'Len of tokens lines not equal'

    for c, token in zip(token_types, parsed_line):
        assert isinstance(token, c)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

from string import ascii_uppercase

import six

__all__ = ('col_str_to_index', 'col_index_to_str', 'u', 'cached_property', 'digit',
           'Array')


def col_str_to_index(col_str):
    """
    A -> 1
    B -> 2
    Z -> 26
    AA -> 27
    :param basestring col_str: [A-Z]+
    :rtype: int
    """
    str_len = len(col_str)
    base = len(ascii_uppercase)
    return sum((ascii_uppercase.index(s) + 1) * base ** (str_len - i)
               for i, s in enumerate(col_str, 1))


def col_index_to_str(i):
    base = len(ascii_uppercase)
    chars = []
    while i:
        i, r = divmod(i, base)
        if r == 0:
            r = base
            i -= 1
        chars.append(ascii_uppercase[r - 1])
    chars.reverse()
    return ''.join(chars)


if six.PY2:
    b_from_default_type = str

    def u_from_default_type(v):
        return str(v).decode('utf8')
else:
    def b_from_default_type(v):
        return str(v).encode('utf8')

    u_from_default_type = str


def u(value):
    if isinstance(value, six.binary_type):
        return value.decode('utf8')
    elif isinstance(value, six.text_type):
        return value
    else:
        return u_from_default_type(value)


class cached_property(object):
    def __init__(self, func):
        self.func = func

    def __get__(self, instance, cls=None):
        result = instance.__dict__[self.func.__name__] = self.func(instance)
        return result


def digit(v):
    if isinstance(v, six.string_types):
        v = float(v)
    elif isinstance(v, bool):
        v = int(v)
    elif v is None:
        v = 0
    return v


class Array(object):
    def __init__(self, array=None):
        self._array = array or []
        self._pos = -1

    def append(self, v):
        self._array.append(v)

    @property
    def is_ended(self):
        return self._pos + 1 >= len(self._array)

    def __next__(self):
        if self.is_ended:
            raise StopIteration()
        else:
            self._pos += 1
            v = self._array[self._pos]
            return v

    def next(self):
        return self.__next__()

    def prev(self):
        return self._array[self._pos - 1] if self._pos > 0 else None

    def current(self):
        return self._array[self._pos] if self._pos >= 0 else None

    def __len__(self):
        return len(self._array)

    def reset(self):
        self._pos = -1

    def __iter__(self):
        return iter(self._array)

    def step_back(self, step=1):
        if step > self._pos + 1:
            self.reset()
        else:
            self._pos -= step

    def __getitem__(self, item):
        return self._array[item]



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

from efc.interface import BaseExcelInterface
from efc.rpn_builder.parser.operands import CellSetOperand, SingleCellOperand


class ExcelMock(BaseExcelInterface):
    data = {
        'Sheet 1': {
            1: {1: 13, 2: 16, 3: 18},
            3: {1: 4, 2: 2, 3: 8},
        },
        'Yet another sheet': {
            100: {1: 4, 2: 2, 3: 8},
            104: {1: 4, 2: 2, 3: 8, 27: 45},
        },
        'Sheet4': {
            1: {1: 13, 2: 16, 3: 18},
            2: {1: 13, 2: 16, 3: 18},
            3: {1: 4, 2: 2, 3: 8},
        },
    }

    def cell_to_value(self, row, column, ws_name):
        return self.data[ws_name].get(row, {}).get(column)

    @property
    def named_ranges(self):
        op_set = CellSetOperand(ws_name='Sheet 1', source=self)
        op_set.add_row([SingleCellOperand(1, 2, ws_name='Sheet 1', source=self),
                        SingleCellOperand(1, 3, ws_name='Sheet 1', source=self)])
        return {
            'test': SingleCellOperand(1, 2, ws_name='Sheet 1', source=self),
            'test2': op_set
        }

    def named_range_to_cells(self, name, ws_name):
        return self.named_ranges[name]

    def has_worksheet(self, ws_name):
        return ws_name in self.data

    def has_named_range(self, name, ws_name):
        return name in self.named_ranges



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

from efc.base.errors import BaseEFCException


class ParserError(BaseEFCException):
    def __init__(self, ws_name, formula):
        self.ws_name = ws_name
        self.formula = formula


class InconsistentParentheses(ParserError):
    code = 200
    msg = 'Formula on ws "{ws_name}" has inconsistent parentheses'


class SeparatorBlockError(ParserError):
    code = 201
    msg = 'Separator without brackets on ws "{ws_name}. Should be like "(op1, op2)"'



# -*- coding: utf8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

from collections import defaultdict

from six import python_2_unicode_compatible, text_type
from six.moves import range

from efc.base.errors import BaseEFCException
from efc.utils import cached_property, col_index_to_str, digit, u

__all__ = ('Operand', 'ErrorOperand', 'ValueErrorOperand', 'WorksheetNotExist',
           'ZeroDivisionErrorOperand', 'SimpleOperand', 'SingleCellOperand',
           'CellSetOperand', 'SimpleSetOperand', 'NamedRangeOperand', 'CellRangeOperand',
           'FunctionNotSupported', 'NotFoundErrorOperand', 'RPNOperand', 'OperandLikeObject', 'OffsetMixin')


class OperandLikeObject(object):
    def __init__(self, ws_name=None, source=None, *args, **kwargs):
        super(OperandLikeObject, self).__init__(*args, **kwargs)
        self.ws_name = ws_name
        self.source = source


@python_2_unicode_compatible
class Operand(OperandLikeObject):
    value = None

    @property
    def digit(self):
        """Digit type"""
        return digit(self.value) if self.value else 0

    @property
    def string(self):
        """String type"""
        if isinstance(self.value, bool):
            return u(self.value).upper()
        elif self.value is not None:
            return u(self.value)
        else:
            return ''

    @property
    def is_blank(self):
        return self.value in (None, '')

    def __int__(self):
        return int(self.value)

    def __float__(self):
        return float(self.value)

    def __str__(self):
        return self.string

    def __trunc__(self):
        return self.__int__()


class ErrorOperand(OperandLikeObject, BaseEFCException):
    code = 300
    msg = 'Unknown error'
    string_value = '#ERROR!'

    def __init__(self, *args, **kwargs):
        self.formula = kwargs.pop('formula', None)
        super(ErrorOperand, self).__init__(*args, **kwargs)

    @property
    def value(self):
        raise self

    @property
    def string(self):
        return self.string_value

    def __getattr__(self, item):
        raise self


class ValueErrorOperand(ErrorOperand):
    code = 301
    msg = 'Cell value error'
    string_value = '#VALUE!'


class WorksheetNotExist(ErrorOperand):
    code = 302
    msg = 'Worksheet does not exist'
    string_value = '#REF!'


class NamedRangeNotExist(ErrorOperand):
    code = 303
    msg = 'Named range "{name}" does not exist'
    string_value = '#NAME?'

    def __init__(self, name, *args, **kwargs):
        super(NamedRangeNotExist, self).__init__(*args, **kwargs)
        self.name = name


class ZeroDivisionErrorOperand(ErrorOperand):
    code = 304
    msg = 'Zero division'
    string_value = '#DIV/0!'


class NotFoundErrorOperand(ErrorOperand):
    code = 305
    msg = 'Result not found'
    string_value = '#VALUE!'


class FunctionNotSupported(ErrorOperand):
    code = 306
    msg = 'Function "{f_name}" not found among available functions'
    string_value = '#NAME?'

    def __init__(self, f_name, *args, **kwargs):
        super(FunctionNotSupported, self).__init__(*args, **kwargs)
        self.f_name = f_name


class SimpleOperand(Operand):
    def __init__(self, value, *args, **kwargs):
        super(SimpleOperand, self).__init__(*args, **kwargs)
        self.value = value


class CellsOperand(OperandLikeObject):
    def address_to_value(self):
        raise NotImplementedError

    @cached_property
    def value(self):
        if self.source.has_worksheet(self.ws_name):
            return self.address_to_value()
        else:
            raise WorksheetNotExist(ws_name=self.ws_name)

    def get_iter(self):
        raise NotImplementedError

    def __iter__(self):
        return self.get_iter()


class OffsetMixin(object):
    def offset(self, row_offset=0, col_offset=0):
        raise NotImplementedError


class SingleCellOperand(CellsOperand, Operand, OffsetMixin):
    def __init__(self, row, column, row_fixed=False, column_fixed=False, *args, **kwargs):
        super(SingleCellOperand, self).__init__(*args, **kwargs)
        self.row = row
        self.column = column
        self.row_fixed = row_fixed
        self.column_fixed = column_fixed

    def address_to_value(self):
        return self.source.cell_to_value(self.row, self.column, self.ws_name)

    def get_iter(self):
        yield self

    @property
    def address(self):
        return "'%s'!%s%d" % (self.ws_name, col_index_to_str(self.column), self.row)

    def offset(self, row_offset=0, col_offset=0):
        new_operand = SingleCellOperand(row=self.row, column=self.column,
                                        row_fixed=self.row_fixed, column_fixed=self.column_fixed,
                                        ws_name=self.ws_name, source=self.source)
        if not new_operand.row_fixed:
            new_operand.row += row_offset
        if not new_operand.column_fixed:
            new_operand.column += col_offset
        return new_operand


class SetOperand(OperandLikeObject):
    operands_type = None

    def __init__(self, *args, **kwargs):
        super(SetOperand, self).__init__(*args, **kwargs)
        self._items = defaultdict(list)

    def check_type(self, items):
        if isinstance(items, list):
            if any(not isinstance(i, self.operands_type) for i in items):
                raise ValueErrorOperand()
        elif not isinstance(items, self.operands_type):
            raise ValueErrorOperand()

    def add_cell(self, item, row=0):
        self.check_type(item)
        self._items[row].append(item)

    def add_many(self, items, row=0):
        self.check_type(items)
        append = self._items[row].append
        for item in items:
            append(item)

    def add_row(self, items):
        self.check_type(items)
        r = max(self._items) + 1 if self._items else 1
        self.add_many(items, r)

    def get_iter(self):
        for r in sorted(self._items):
            for item in self._items[r]:
                yield item

    def __iter__(self):
        return self.get_iter()

    @cached_property
    def value(self):
        return list(self)


class CellSetOperand(SetOperand):
    operands_type = SingleCellOperand


class SimpleSetOperand(SetOperand):
    operands_type = SimpleOperand


class CellRangeOperand(CellsOperand, OffsetMixin):
    def __init__(self, row1, column1, row2, column2,
                 row1_fixed=False, column1_fixed=False, row2_fixed=False, column2_fixed=False,
                 *args, **kwargs):
        super(CellRangeOperand, self).__init__(*args, **kwargs)
        self.row1 = row1
        self.column1 = column1
        self.row2 = row2
        self.column2 = column2

        self.row1_fixed = row1_fixed
        self.column1_fixed = column1_fixed
        self.row2_fixed = row2_fixed
        self.column2_fixed = column2_fixed

    def get_iter(self):
        column1 = self.source.min_column(self.ws_name) if self.column1 is None else self.column1
        column2 = self.source.max_column(self.ws_name) if self.column2 is None else self.column2
        row1 = self.source.min_row(self.ws_name) if self.row1 is None else self.row1
        row2 = self.source.max_row(self.ws_name) if self.row2 is None else self.row2

        if row1 == row2:
            for c in range(column1, column2 + 1):
                yield SingleCellOperand(row1, c, ws_name=self.ws_name, source=self.source)
        elif column1 == column2:
            for r in range(row1, row2 + 1):
                yield SingleCellOperand(r, column1, ws_name=self.ws_name, source=self.source)
        else:
            for r in range(row1, row2 + 1):
                for c in range(column1, column2 + 1):
                    yield SingleCellOperand(r, c, ws_name=self.ws_name, source=self.source)

    def address_to_value(self):
        return self.get_iter()

    @property
    def address(self):
        return "'%s'!%s%d:%s%d" % (self.ws_name,
                                   col_index_to_str(self.column1), self.row1,
                                   col_index_to_str(self.column2), self.row2)

    def offset(self, row_offset=0, col_offset=0):
        new_operand = CellRangeOperand(row1=self.row1, column1=self.column1,
                                       row2=self.row2, column2=self.column2,
                                       row1_fixed=self.row1_fixed, column1_fixed=self.column1_fixed,
                                       row2_fixed=self.row2_fixed, column2_fixed=self.column2_fixed,
                                       ws_name=self.ws_name, source=self.source)

        if not new_operand.row1_fixed and new_operand.row1 is not None:
            new_operand.row1 += row_offset
        if not new_operand.column1_fixed and new_operand.column1 is not None:
            new_operand.column1 += col_offset
        if not new_operand.row2_fixed and new_operand.row2 is not None:
            new_operand.row2 += row_offset
        if not new_operand.column2_fixed and new_operand.column2 is not None:
            new_operand.column2 += col_offset
        return new_operand


class NamedRangeOperand(CellsOperand):
    def __init__(self, name, *args, **kwargs):
        super(NamedRangeOperand, self).__init__(*args, **kwargs)
        self.name = name

    def address_to_value(self):
        return self.source.named_range_to_cells(self.name, self.ws_name)

    @cached_property
    def value(self):
        if not self.source.has_worksheet(self.ws_name):
            raise WorksheetNotExist(ws_name=self.ws_name)
        elif not self.source.has_named_range(self.name, self.ws_name):
            raise NamedRangeNotExist(self.name, self.ws_name)
        else:
            return self.address_to_value()

    def get_iter(self):
        return iter(self.value)


@python_2_unicode_compatible
class RPNOperand(OperandLikeObject, OffsetMixin):
    def __init__(self, rpn, *args, **kwargs):
        super(RPNOperand, self).__init__(*args, **kwargs)
        self.rpn = rpn
        self._result = None

    @cached_property
    def evaluated_value(self):
        v = self.rpn.calc(ws_name=self.ws_name, source=self.source)
        if isinstance(v, RPNOperand):
            v = v.evaluated_value
        return v

    def __getattr__(self, item):
        return getattr(self.evaluated_value, item)

    def offset(self, row_offset=0, col_offset=0):
        return RPNOperand(rpn=self.rpn.offset(row_offset, col_offset), ws_name=self.ws_name, source=self.source)

    def __int__(self):
        return int(self.evaluated_value)

    def __float__(self):
        return float(self.evaluated_value)

    def __str__(self):
        return text_type(self.evaluated_value)

    def __trunc__(self):
        return self.__int__()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

from functools import wraps

from efc.rpn_builder.parser.functions import EXCEL_FUNCTIONS
from efc.rpn_builder.parser.operands import FunctionNotSupported, OperandLikeObject, SimpleOperand


def excel_function(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        result = func(*args, **kwargs)
        if not isinstance(result, OperandLikeObject):
            result = SimpleOperand(result)
        return result

    return wrapper


class Operation(object):
    operands_count = None
    priority = None

    def __init__(self, f_name):
        self.f_name = f_name
        self.operands_count = 1

    @property
    def f(self):
        return excel_function(EXCEL_FUNCTIONS[self.f_name])

    @property
    def is_exists(self):
        return self.f_name in EXCEL_FUNCTIONS

    def eval(self, *args):
        if self.is_exists:
            return self.f(*args)
        else:
            return FunctionNotSupported(self.f_name)


class ArithmeticOperation(Operation):
    def __init__(self, f_name, priority):
        super(ArithmeticOperation, self).__init__(f_name)
        self.operands_count = 2
        self.priority = priority


class FunctionOperation(Operation):
    def __init__(self, f_name):
        super(FunctionOperation, self).__init__(f_name)
        self.operands_count = 1
        self.priority = float('inf')



# -*- coding: utf-8 -*-

from __future__ import unicode_literals, print_function
from efc.utils import u
import six


@six.python_2_unicode_compatible
class BaseEFCException(Exception):
    code = 0
    msg = None

    def __str__(self):
        context = {k: u(i) for k, i in six.iteritems(self.__dict__) if i is not None}
        msg_list = []
        if self.code is not None:
            msg_list.append('Code %d' % self.code)
        if self.msg:
            msg_list.append(self.msg.format(**context))
        if self.__dict__.get('formula'):
            msg_list.append('Formula: %s' % self.__dict__['formula'])
        if self.__dict__.get('ws_name'):
            msg_list.append('WS: %s' % self.__dict__['ws_name'])
        return '. '.join(msg_list)
